"""Набор инструментов адаптера."""

import logging

from .adapters import Adapter


logger = logging.getLogger(__name__)


def cleanup(adapter: Adapter):

    """Полная очистка.

    Выбирает и удаляет все развёртывания и определения процессов и их активные задачи.
    """

    logger.info('Cleanup: removing definitions')
    for definition in adapter.get_definitions():
        logger.info('Removing %s', definition)
        adapter.rm_definition(definition, cascade=True)

    logger.info('Cleanup: removing deployments')
    for deployment in adapter.get_deployments():
        logger.info('Removing %s', deployment)
        adapter.rm_deployment(deployment, cascade=True)

    assert not list(adapter.get_deployments())
    assert not list(adapter.get_definitions())

    logger.info('Cleanup: done')



"""Набор тестов для проверки адаптера к Camunda."""

from pathlib import Path
from typing import List
import logging
import unittest

from more_itertools import consume, ilen
import requests

from testapp import core
from testapp.core.adapters.db import comissions_repository as comissions
from testapp.core.adapters.db import declarations_repository as declarations
from testapp.core.adapters.db import notifications_repository as notifications
from testapp.core.adapters.db import sessions_repository as sessions
from testapp.core.domain.model import STATUS_PENDING, Declaration

from camunda_adapter import Adapter, Config
from camunda_adapter.adapters import TopicHandlerRegistry
from camunda_adapter.domain import model
from camunda_adapter.utils import cleanup


logger = logging.getLogger(__name__)


PROCESS_DEFINITION_ID = 'camunda_adapter_test'

DECLARATIONS_COUNT = 100

logger.info('PROCESS_DEFINITION_ID: %s', PROCESS_DEFINITION_ID)

assert core.adapter is not None


class TestAdapter(unittest.TestCase):

    """Набор тестов для проверки адаптера к Camunda."""

    config: Config
    adapter: Adapter
    definition: model.Definition

    @classmethod
    def _setup_deployment(cls) -> None:

        assert core.adapter is not None

        cls.adapter = core.adapter

        cleanup(cls.adapter)
        try:
            cls.definition = cls.adapter.get_definition_by_key(PROCESS_DEFINITION_ID)
            logger.info('Definition exists')
        except ValueError:
            cls.definition = cls.adapter.create_deployment(
                PROCESS_DEFINITION_ID,
                Path(__file__).parent / 'diagrams' / f'{PROCESS_DEFINITION_ID}.bpmn'
            )
            logger.info('Definition deployed')

    @classmethod
    def setUpClass(cls):
        try:
            cls._setup_deployment()
        except requests.exceptions.RequestException as e:
            logger.exception('Can not deploy Camunda')
            response = getattr(e, 'response', None)
            if response is not None:
                error = model.Error(**response.json())
                logger.error(error)
            raise

    @classmethod
    def tearDownClass(cls):
        """Очистка."""
        cleanup(cls.adapter)

    # @debug_requests
    def test_camunda_adapter(self) -> None:
        assert isinstance(self.adapter, Adapter)

        consume(map(declarations.add, (Declaration() for _ in range(DECLARATIONS_COUNT))))

        # Все ID уникальны
        unique_ids_count = len(set(d.id for d in declarations.get_all_objects()))
        self.assertEqual(unique_ids_count, DECLARATIONS_COUNT)

        # Выбор процесса для запуска
        definition: model.Definition = self.adapter.get_definition_by_key(PROCESS_DEFINITION_ID)
        self.assertIsInstance(definition, model.Definition)

        def send_messages(*declarations_: Declaration) -> None:
            for declaration in declarations_:
                with self.subTest(declaration):
                    message = model.Message(
                        messageName='DeclarationReceived',
                        businessKey=f'attestation-declaration-{declaration.id}',
                        processVariables={
                            name: model.Variable(value=value)
                            for name, value in declaration.dict(include={'id'}).items()
                        }
                    )
                    message_results: List[model.MessageResult] = list(self.adapter.message(message))
                    self.assertEqual(len(message_results), 1)

                    process = message_results[0].processInstance

                    assert process is not None

                    self.assertIsInstance(process, model.Process)

                    self.assertIn(PROCESS_DEFINITION_ID, process.definitionId)
                    self.assertIsNotNone(process.id)

                    processes = list(self.adapter.get_processes())
                    self.assertGreater(len(processes), 0)

                    self.assertIn(process, processes)

                    # Проверка наличия внешних (по отн. к Camunda) задач
                    external_tasks = list(self.adapter.get_external_tasks(definition=definition))
                    self.assertGreater(len(external_tasks), 0)

        # Запуск процесса через сообщение
        send_messages(*declarations.get_all_objects())

        while ilen(self.adapter.get_external_tasks(definition=definition)):
            consume(self.adapter.process_external_tasks())

        def get_count(attr):
            """Количество заявлений с attr=True."""
            return ilen(
                filter(
                    lambda x: bool(getattr(x, attr, None)) is True,
                    declarations.get_all_objects()
                )
            )

        # Соответствие состояния объекта его статусу
        # означает прохождение всех переходов модели.
        for declaration in declarations.get_all_objects():
            with self.subTest(declaration):

                self.assertNotEqual(declaration.status, STATUS_PENDING)

                if declaration.rejected:
                    self.assertEqual(declaration.status, 'Отказано')

                if declaration.withdrawn:
                    self.assertEqual(declaration.status, 'Отозвано')

                if declaration.needs_work:
                    self.assertIn(declaration.status, ['На доработку', 'Отозвано'])

                if not any((declaration.rejected, declaration.withdrawn, declaration.needs_work)):
                    self.assertIn(declaration.status, ['Рассмотрено'])

        # Добавлены уведомления для заявлений с доработками
        self.assertGreaterEqual(
            ilen(notifications.get_all_objects()),
            get_count('needs_work')
        )

        # Проведены заседания и комиссии для рассмотренных заявлений
        for declaration in filter(
            lambda declaration: declaration.status == 'Рассмотрено',
            declarations.get_all_objects()
        ):
            with self.subTest(declaration):
                self.assertIsNotNone(sessions.get_by_declaration(declaration))
                self.assertIsNotNone(comissions.get_by_declaration(declaration))


class TestTopicRegistry(unittest.TestCase):

    """Набор тестов для проверки реестра обработчиков."""

    def test_topic_registry(self):

        def handler1(task, adapter):  # pylint:disable=unused-argument
            pass

        def handler2(task, adapter):  # pylint:disable=unused-argument
            pass

        topic = model.Topic(name='test-topic')

        registry = TopicHandlerRegistry()

        registry.register(topic, handler1)
        self.assertIs(handler1, registry.get(topic))

        registry.replace(topic, handler2)
        self.assertIs(handler2, registry.get(topic))

        not_registered_topic = model.Topic(name='not-registered-topic')
        with self.assertRaises(ValueError):
            registry.get(not_registered_topic)

        occasionally_same_topic = model.Topic(name='test-topic')
        with self.assertRaises(ValueError):
            registry.register(occasionally_same_topic, handler1)



"""Модель ПО "Заявления на аттестацию"."""

from typing import List, Union
from uuid import uuid4

from pydantic.fields import Field
from pydantic.main import BaseModel


STATUS_PENDING = 'В очереди'


class Declaration(BaseModel):

    """Заявление."""

    id: str = Field(
        title='Уникальный ID',
        default_factory=lambda: str(uuid4())
    )
    stage: Union[str, None] = Field(
        title='Этап', default=None
    )
    status: str = Field(
        title='Статус', default=STATUS_PENDING
    )
    approved: Union[bool, None] = Field(
        title='Одобрено секретарём', default=False
    )
    rejected: Union[bool, None] = Field(
        title='Отклонено секретарём', default=False
    )
    needs_work: Union[bool, None] = Field(
        title='Требует доработки', default=False
    )
    withdrawn: Union[bool, None] = Field(
        title='Отозвано', default=False
    )
    experts: List[int] = Field(
        title='Эксперты', default_factory=list
    )


class Session(BaseModel):

    """Заседание."""

    declaration: Declaration = Field(title='Заявление')


class Comission(BaseModel):

    """Комиссия."""

    declaration: Declaration = Field(title='Заявление')


class Notification(BaseModel):

    """Уведомление."""

    declaration: Declaration = Field(title='Заявление')



"""Модель взаимодействия с платформой Camunda."""

# pylint: disable=C0103

from datetime import date
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic.main import BaseModel
from pydantic.networks import AnyUrl


TYPE_MAP = {
    str: 'String',
    bool: 'Boolean',
    int: 'Integer',
    float: 'Double',
    date: 'Date',
    type(None): 'Null'
}

REVERSE_MAP = {
    val: key
    for (key, val) in TYPE_MAP.items()
}


class Deployment(BaseModel):

    """Данные о развёртывании процесса."""

    id: str
    name: Optional[str] = None
    source: Optional[str] = None
    tenantId: Optional[str] = None
    deploymentTime: str

    deployedProcessDefinitions: Optional[Dict[str, 'Definition']] = {}
    deployedDecisionDefinitions: Optional[Dict[str, Any]] = {}
    deployedDecisionRequirementsDefinitions: Optional[Dict[str, Any]] = {}


class Definition(BaseModel):

    """Определение процесса."""

    key: str
    id: str
    name: str
    deploymentId: UUID
    version: int

    startableInTasklist: bool = False
    suspended: bool = False
    tenantId: Optional[str] = None

    versionTag: Optional[str] = None
    description: Optional[str] = None
    resource: Optional[str] = None
    category: AnyUrl
    diagram: Optional[str] = None
    historyTimeToLive: Optional[int] = None


class Process(BaseModel):

    """Запущенный процесс."""

    id: str
    definitionId: str
    businessKey: Optional[str] = None
    caseInstanceId: Optional[str] = None
    ended: bool = False
    suspended: bool = False
    tenantId: Optional[str] = None


class Topic(BaseModel):

    """Топик для подписки обработчиков."""

    name: str

    class Config:
        frozen = True


class Variable(BaseModel):

    """Переменная задачи, сообщения."""

    type: str
    value: Union[str, bool, int, float, date, None]  # См. TYPE_MAP
    valueInfo: dict = {}

    def __init__(self, *args, **kwargs):
        kwargs['type'] = TYPE_MAP[type(kwargs['value'])]
        super().__init__(*args, **kwargs)


Variables = Dict[str, Variable]
"""Переменные задач, сообщений."""


class ExternalTask(BaseModel):

    """Внешняя задача.

    Считается внешней по отн. к Camunda, исполняется воркером или
    отдельным приложением.
    """

    id: str
    topicName: str
    businessKey: Optional[str] = None

    processDefinitionId: str
    processDefinitionKey: str
    processDefinitionVersionTag: Optional[str] = None

    processInstanceId: str

    activityId: str
    activityInstanceId: str
    executionId: str

    variables: Variables = {}

    errorMessage: Optional[str] = None

    lockExpirationTime: Optional[str] = None
    retries: Optional[int] = None
    suspended: bool = False
    workerId: Optional[str] = None
    tenantId: Optional[str] = None
    priority: Optional[int] = None
    extensionProperties: dict = {}

    topic: Topic

    def __init__(self, *args, **kwargs):
        topic = Topic(name=kwargs['topicName'])
        super().__init__(*args, topic=topic, **kwargs)


class Message(BaseModel):

    """Сообщение.

    Может использоваться как возникающее событие.
    """

    messageName: str
    businessKey: str
    resultEnabled: Optional[bool] = True
    variablesInResultEnabled: Optional[bool] = True
    correlationKeys: Variables = {}
    processVariables: Variables = {}


class Execution(BaseModel):

    """Среда исполнения процесса."""

    id: str
    processInstanceId: str
    ended: bool
    tenantId: Union[str, None] = None


class MessageResult(BaseModel):

    """Результат отправки сообщения."""

    resultType: str
    execution: Optional[Union[Execution, None]] = None
    processInstance: Optional[Process] = None
    variables: Variables = {}


class ErrorDetail(BaseModel):

    """Детальная информация об ошибке."""

    column: int
    line: int
    mainElementId: str
    message: str
    elementIds: List[str]


TWarning = Any


class MessageDetail(BaseModel):

    """Детальная информация сообщения."""

    errors: List[ErrorDetail]
    warnings: List[TWarning]


TMessageDetails = Dict[str, MessageDetail]


class Error(BaseModel):

    """Ошибка запроса."""

    type: str
    message: str
    code: int = 0
    details: Optional[TMessageDetails] = None


Deployment.update_forward_refs()



"""Обработчики внешних задач."""

import logging
import random

from more_itertools.recipes import consume

from testapp.core.adapters.db import comissions_repository as comissions
from testapp.core.adapters.db import declarations_repository as declarations
from testapp.core.adapters.db import notifications_repository as notifications
from testapp.core.adapters.db import sessions_repository as sessions
from testapp.core.domain import model

from camunda_adapter.adapters import Adapter
from camunda_adapter.domain import model as camunda


logger = logging.getLogger(__name__)


set_stage_topic = camunda.Topic(name='set-stage')


def set_stage(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)
    assert isinstance(task.variables['stage'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    declaration.stage = task.variables['stage'].value
    logger.info(
        'Установлен этап "%s" заявлению "%s"',
        task.variables['stage'].value, declaration.id
    )

    return adapter.complete_external_task(task)


set_status_topic = camunda.Topic(name='set-status')


def set_status(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)
    assert isinstance(task.variables['status'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    declaration.status = task.variables['status'].value

    adapter.complete_external_task(task)

    logger.info(
        'Установлен статус "%s" заявлению "%s"',
        task.variables['status'].value, declaration.id
    )


def get_bool(from_=(True, False)):
    return bool(random.choice(from_))  # noqa


def imitate_secretary(declaration: model.Declaration):
    """Рандомно отклоняем, возвращаем или принимаем заявление."""

    # Для вернувшихся заявлений сбрасываем результаты проверки
    declaration.approved = declaration.rejected = declaration.needs_work = None

    declaration.rejected = not get_bool(from_=range(10))  # ~10%

    if not declaration.rejected:
        declaration.needs_work = not get_bool(from_=range(10))  # ~10%

    declaration.approved = (
        not declaration.rejected and
        not declaration.needs_work
    )


def declaration_withdrawn(declaration):
    declaration.withdrawn = not get_bool(from_=range(20))  # ~5% за проверку
    logger.info('Отозвано заявление "%s"', declaration.id)


def check_withdrawn(declaration, adapter: Adapter):
    declaration_withdrawn(declaration)
    if declaration.withdrawn:
        message = camunda.Message(
            messageName='DeclarationWithdrawn',
            businessKey=f'attestation-declaration-{declaration.id}',
            processVariables={
                name: camunda.Variable(value=value)
                for name, value in declaration.dict(include={'id'}).items()
            }
        )
        consume(adapter.message(message))
    return declaration.withdrawn


approve_declaration_topic = camunda.Topic(name='approve-declaration')


def approve_declaration(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    if check_withdrawn(declaration, adapter):
        return

    imitate_secretary(declaration)

    logger.info(
        (
            'Результаты проверки секретарём: '
            'принято: %s, отклонено: %s, на доработку: %s'
        ),
        declaration.approved, declaration.rejected, declaration.needs_work
    )
    return adapter.complete_external_task(
        task,
        approved=declaration.approved,
        rejected=declaration.rejected,
        needs_work=declaration.needs_work
    )


get_withdrawn_topic = camunda.Topic(name='get-withdrawn')


def get_withdrawn(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    # Небольшая вероятность отзыва заявления
    if check_withdrawn(declaration, adapter):
        return

    return adapter.complete_external_task(task, withdrawn=declaration.withdrawn)


set_experts_topic = camunda.Topic(name='set-experts')


def set_experts(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    available_experts = list(range(1, 10))

    if check_withdrawn(declaration, adapter):
        return

    if not get_bool(from_=range(20)):
        logger.info('Ожидание назначения экспертов: %s', task.id)
        return

    declaration.experts = list(set(random.choice(available_experts) for _ in range(7)))
    logger.info('Назначаются эксперты: %s', declaration.experts)
    return adapter.complete_external_task(task, expert_ids=declaration.experts)


notify_needs_work_topic = camunda.Topic(name='notify-needs-work')


def notify_needs_work(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    logger.info('Уведомление о доработке заявления "%s"', declaration.id)
    notifications.add(model.Notification(declaration=declaration))
    return adapter.complete_external_task(task)


expertise_expert_validate_declaration_topic = camunda.Topic(name='expertise-expert-validate-declaration')


def expertise_expert_validate_declaration(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    expert_id: camunda.Variable = task.variables['expert_id']
    assert expert_id.type in camunda.REVERSE_MAP

    logger.info('Ознакомление: эксперт %s, заявление %s, задача %s', expert_id.value, declaration.id, task.id)

    assert camunda.REVERSE_MAP[expert_id.type](expert_id.value), declaration.experts

    return adapter.complete_external_task(task, expertise_declaration_valid=get_bool())


expertise_expert_do_expertise_topic = camunda.Topic(name='expertise-expert-do-expertise')


def expertise_expert_do_expertise(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    expert_id: camunda.Variable = task.variables['expert_id']
    assert expert_id.type in camunda.REVERSE_MAP

    logger.info('Экспертиза: эксперт %s, заявление %s, задача %s', expert_id.value, declaration.id, task.id)

    assert camunda.REVERSE_MAP[expert_id.type](expert_id.value), declaration.experts

    return adapter.complete_external_task(task, expertise_declaration_valid=get_bool())


set_session_topic = camunda.Topic(name='set-session')


def set_session(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    logger.info('Назначение сессии "%s"', declaration.id)
    sessions.add(model.Session(declaration=declaration))
    return adapter.complete_external_task(task)


set_comission_results_topic = camunda.Topic(name='set-comission-results')


def set_comission_results(task: camunda.ExternalTask, adapter: Adapter):
    assert isinstance(task.variables['id'].value, str)

    declaration = declarations.get_by_id(task.variables['id'].value)
    assert declaration is not None

    logger.info('Назначение результатов комиссии "%s"', declaration.id)
    comissions.add(model.Comission(declaration=declaration))
    return adapter.complete_external_task(task)



"""Ядро пиложения."""

from typing import Union

from camunda_adapter.adapters import Adapter
from camunda_adapter.config import Config


# Предоставляются при инициализации
config: Union[Config, None] = None
adapter: Union[Adapter, None] = None



"""In-memory репозитории, имитирующие хранение данных."""

from abc import ABC
from typing import Any, Iterable, List, Union

from testapp.core.domain import model


class AbstractRepository(ABC):

    """Абстрактный репозиторий."""

    _items: Any

    def __init__(self) -> None:
        self._items = []

    def add(self, item: Any) -> None:
        self._items.append(item)

    def get_all_objects(self) -> Iterable[Any]:
        yield from iter(self._items)


class DeclarationsRepository(AbstractRepository):

    """Хранилище заявлений."""

    _items: List[model.Declaration]

    def add(self, item: model.Declaration) -> None:
        assert isinstance(item, model.Declaration)
        return super().add(item)

    def get_all_objects(self) -> Iterable[model.Declaration]:
        return super().get_all_objects()

    def get_by_id(self, id_: str) -> Union[model.Declaration, None]:
        return next(filter(lambda d: d.id == id_, self._items))


declarations_repository = DeclarationsRepository()


class SearchByDeclarationMixin(AbstractRepository):

    """Примесь к хранилищу для поиска по заявлению."""

    def get_by_declaration(
        self, declaration: model.Declaration
    ) -> Union[model.Session, None]:
        return next(
            filter(
                lambda item: item.declaration == declaration,
                self._items
            )
        )


class SessionsRepository(SearchByDeclarationMixin, AbstractRepository):

    """Хранилище заседаний."""

    _items: List[model.Session]


sessions_repository = SessionsRepository()


class NotificationsRepository(SearchByDeclarationMixin, AbstractRepository):

    """Хранилище уведомлений."""

    _items: List[model.Notification]


notifications_repository = NotificationsRepository()


class ComissionsRepository(SearchByDeclarationMixin, AbstractRepository):

    """Хранилище комиссий."""

    _items: List[model.Comission]


comissions_repository = ComissionsRepository()



"""Инструменты конфигурации адаптера."""

from typing import TYPE_CHECKING
from uuid import uuid4

from pydantic.fields import Field
from pydantic.networks import HttpUrl


if TYPE_CHECKING:
    from dataclasses import dataclass  # noqa
else:
    from pydantic.dataclasses import dataclass  # noqa


BPMN_ENGINE_URL = 'BPMN_ENGINE_URL'
BPMN_PROCESS_ID = 'BPMN_PROCESS_ID'
BPMN_LOCK_DURATION = 'BPMN_LOCK_DURATION'
BPMN_ENGINE_REQUEST_TIMEOUT = 'BPMN_ENGINE_REQUEST_TIMEOUT'
BPMN_WORKER_ID = 'BPMN_WORKER_ID'


DEFAULT_REQUEST_TIMEOUT = 30.0  # сек


@dataclass  # pylint: disable=used-before-assignment
class Config:

    """Конфигурация адаптера."""

    engine_url: HttpUrl = Field(
        title='URL движка',
        default='http://127.0.0.1:8080/engine-rest/'
    )
    lock_duration: int = Field(
        title='Продолжительность блокировки задачи при её получении',
        default=1000  # мс
    )
    request_timeout: float = Field(
        title='Таймаут запроса к BPMN',
        default=DEFAULT_REQUEST_TIMEOUT
    )
    worker_id: str = Field(
        title='ID воркера (для блокировки задач)',
        default=f'camunda-adapter-{str(uuid4())}'
    )



"""Bootstrap-модуль для настройки приложения."""

# pylint: disable=import-outside-toplevel

from django.apps import AppConfig as AppConfigBase

from camunda_adapter.adapters import Adapter
from camunda_adapter.config import Config


class AppConfig(AppConfigBase):

    """Подключение django-app к django-project."""

    name = __package__

    def ready(self):
        self._setup_bpmn()

    def _setup_bpmn(self):
        from testapp import core

        from . import services

        core.config = Config()
        core.adapter = Adapter(core.config)

        for topic, handler in (
            (services.set_stage_topic, services.set_stage),
            (services.set_status_topic, services.set_status),
            (services.approve_declaration_topic, services.approve_declaration),
            (services.get_withdrawn_topic, services.get_withdrawn),
            (services.set_experts_topic, services.set_experts),
            (services.notify_needs_work_topic, services.notify_needs_work),
            (services.expertise_expert_validate_declaration_topic, services.expertise_expert_validate_declaration),
            (services.expertise_expert_do_expertise_topic, services.expertise_expert_do_expertise),
            (services.set_session_topic, services.set_session),
            (services.set_comission_results_topic, services.set_comission_results)
        ):
            core.adapter.register_topic_handler(topic, handler)



# coding: utf-8
from pathlib import Path

from setuptools import find_packages, setup


setup(
    name='camunda-adapter',
    url='https://stash.bars-open.ru/projects/M3/repos/camunda-adapter',
    license='MIT',
    author='BARS Group',
    description='Адаптер для BPMN-движка Camunda',
    author_email='bars@bars-open.ru',
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=('testapp', 'testapp.*',)),
    long_description=(Path(__file__).parent / 'README.md').open('r').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    package_data = {
        '': ['LICENSE', 'README.md5', 'RELEASE']
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Natural Language :: Russian',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Development Status :: 5 - Production/Stable',
    ],
    install_requires=(
        'pydantic',
        'requests'
    ),
    dependency_links=(
        'http://pypi.bars-open.ru/simple/m3-builder',
    ),
    setup_requires=(
        'm3-builder>=1.2,<2',
    ),
    python_requires='>=3.7',
    set_build_info=Path(__file__).parent.as_posix(),
)



"""Модуль адаптера к платформе Camunda."""

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse
from uuid import uuid4
import logging

import requests

from .config import Config
from .const import PATH_SEP
from .domain import model


logger = logging.getLogger(__name__)


TopicHandler = Callable[[model.ExternalTask, 'Adapter'], Any]


class AbstractTopicHandlerRegistry(ABC):

    """Абстрактный реестр обработчиков внешних задач."""

    _registry: Dict[model.Topic, TopicHandler]

    def __init__(self):
        self._registry = {}

    @abstractmethod
    def register(self, topic: model.Topic, handler: TopicHandler):
        """Зарегистрировать обработчик."""

    @abstractmethod
    def replace(self, topic: model.Topic, handler: TopicHandler):
        """Заменить обработчик."""

    @abstractmethod
    def get(self, topic: model.Topic) -> TopicHandler:
        """Получить обработчик по топику."""


class TopicHandlerRegistry(AbstractTopicHandlerRegistry):

    """Реестр обработчиков внешних задач."""

    def register(self, topic: model.Topic, handler: TopicHandler):
        if topic in self._registry:
            raise ValueError(f'Обработчик "{topic}" уже зарегистрирован')
        self._registry[topic] = handler
        logger.info('Обработчик "%s" зарегистрирован', topic)

    def replace(self, topic: model.Topic, handler: TopicHandler):
        if topic not in self._registry:
            logger.warning('Обработчик "%s" не был зарегистрирован', topic)
        self._registry[topic] = handler
        logger.info('Обработчик "%s" зарегистрирован', topic)

    def get(self, topic: model.Topic) -> TopicHandler:
        handler = self._registry.get(topic)
        if handler is None:
            raise ValueError(f'Обработчик "{topic}" не был зарегистрирован')
        return handler


class Session(requests.Session):

    """Сессия, определяющая параметры запросов к Camunda."""

    _config: Config

    def __init__(
        self,
        config: Config,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._config = config

    def request(self, *args, **kwargs):
        kwargs.setdefault('timeout', self._config.request_timeout)
        return super().request(*args, **kwargs)


class Adapter:

    """Адаптер к движку Camunda."""

    _config: Config
    _session: Session

    def __init__(self, config: Config):
        self._config = config
        self._session = Session(config=self._config)
        self._topic_registry = TopicHandlerRegistry()

        logger.info('Camunda adapter worker_id: %s', self._config.worker_id)

    # ########################################################################
    # Deployments
    # ########################################################################
    def get_deployments(
        self
    ) -> Iterable[model.Deployment]:
        """Получить все развертывания."""

        response = self._session.get(
            self._build_url('/deployment')
        )
        response.raise_for_status()
        for data in response.json():
            yield model.Deployment(**data)

    def create_deployment(self, name: str, diagram_file: Path):
        """Создать развертывание.

        Загружает файл BPMN-диаграммы и создает развертывание с именем name.
        """

        with diagram_file.open('r') as fd:
            response = self._session.post(
                self._build_url(
                    '/deployment/create'
                ),
                files={
                    'data': fd
                },
                data={
                    'deployment-name': name,
                    'deploy-changed-only': 'true'
                }
            )
            response.raise_for_status()
            return model.Deployment(**response.json())

    def rm_deployment(
        self,
        deployment: model.Deployment,
        cascade=False
    ):
        """Удалить развертывание."""

        response = self._session.delete(
            self._build_url(
                '/deployment/{id}', cascade=cascade
            ).format(id=deployment.id)
        )
        response.raise_for_status()

    # ########################################################################
    # Definitions
    # ########################################################################
    def get_definitions(
        self
    ) -> Iterable[model.Definition]:
        """Получить все определения процессов."""

        response = self._session.get(
            self._build_url('/process-definition/')
        )
        response.raise_for_status()
        for data in response.json():
            yield model.Definition(**data)

    def get_definition_by_key(
        self,
        key: str
    ) -> model.Definition:
        """Получить определение процесса по его ключу."""

        for definition in self.get_definitions():
            if definition.key == key:
                return definition
        else:
            raise ValueError(f'Процесс "{key}" не найден')

    def rm_definition(
        self,
        process_definition: model.Definition,
        cascade=True,
        skip_custom_listeners=False,
        skip_io_mappings=False
    ):
        """Удалить определение процесса."""

        response = self._session.delete(
            self._build_url(
                '/process-definition/key/{key}',
                cascade=cascade,
                skipCustomListeners=skip_custom_listeners,
                skipIoMappings=skip_io_mappings
            ).format(key=process_definition.key)
        )
        response.raise_for_status()

    # ########################################################################
    # Processes
    # ########################################################################
    def get_processes(self):
        """Получить все экземпляры процессов."""

        response = self._session.get(
            self._build_url('/process-instance')
        )
        response.raise_for_status()
        for data in response.json():
            yield model.Process(**data)

    def start_process(
        self,
        process_definition: model.Definition,
        business_key=None,
        **variables
    ):
        """Запуск процесса.

        Может быть запущен процесс, имеющий StartEvent в качестве точки входа.
        """

        payload = {
            'businessKey': business_key or str(uuid4()),
            'variables': {
                name: {
                    'value': value,
                    'type': model.TYPE_MAP[type(value)]
                } for name, value in variables.items()
            }
        }

        response = self._session.post(
            self._build_url(
                '/process-definition/key/{key}/start'
            ).format(key=process_definition.key),
            json=payload
        )
        response.raise_for_status()
        return model.Process(**response.json())

    # ########################################################################
    # Messages
    # ########################################################################
    def message(
        self, message: model.Message
    ):
        """Передача сообщения процессу с коррелирующим message.businessKey."""

        response = self._session.post(
            self._build_url(
                '/message'
            ),
            json=message.dict()
        )
        response.raise_for_status()
        for data in response.json():
            yield model.MessageResult(**data)

    # ########################################################################
    # External tasks
    # ########################################################################
    def get_external_tasks_topics(self):
        """Получить все топики внешних задач."""

        response = self._session.get(
            self._build_url('/external-task/topic-names')
        )
        for topic_name in response.json():
            yield model.Topic(name=topic_name)

    def get_external_tasks(
        self,
        definition: Optional[model.Definition] = None
    ) -> Iterable[model.ExternalTask]:
        """Запрос списка внешних задач.

        Если указан definition, отдает список задач конкретного типа процессов.
        """

        query_params = {}
        if definition is not None:
            query_params['processDefinitionId'] = definition.id

        response = self._session.get(
            self._build_url('/external-task', **query_params)
        )
        response.raise_for_status()

        logger.info(
            'Получено %s внешних задач (definition=%s)',
            len(response.json()), definition.name if definition else definition
        )

        for data in response.json():
            logger.debug(data)
            yield model.ExternalTask(**data)

    def fetch_and_lock_external_tasks(
        self,
        *topics: model.Topic,
        max_tasks: int = 10,
    ) -> Iterable[model.ExternalTask]:
        """Блокировка и получение списка внешних задач по топикам.

        Задачи, заблокированые ранее, не отдаются.

        max_tasks - ограничение по одновременному количеству задач.
        """

        if not topics:
            yield from ()
            return

        response = self._session.post(
            self._build_url(
                '/external-task/fetchAndLock'
            ),
            json={
                'workerId': self._config.worker_id,
                'maxTasks': max_tasks,
                'usePriority': True,
                'topics': [
                    {
                        'topicName': topic.name,
                        'lockDuration': self._config.lock_duration
                    }
                    for topic in topics
                ]
            }
        )

        response.raise_for_status()

        logger.info('Получено и заблокировано %s задач', len(response.json()))

        for data in response.json():
            logger.debug(data)
            yield model.ExternalTask(**data)

    def complete_external_task(
        self,
        task: model.ExternalTask,
        **variables
    ) -> None:
        """Отметить задачу выполненной.

        variables - новые переменные, которые будут установлены процессу.
        """

        response = self._session.post(
            self._build_url(
                f'/external-task/{task.id}/complete'
            ),
            json={
                'workerId': self._config.worker_id,
                'variables': {
                    name: {'value': value}
                    for name, value in variables.items()
                }
            }
        )
        response.raise_for_status()

    def _build_url(self, *paths: str, **query_params) -> str:
        """Строит URL запроса исходя из параметров подключения и переданных значений.

        ... code:: python
          >>> self._build_url('/api/', '/v1/', cascade=True)

              http://127.0.0.1:8080/api/v1/?cascade=true
        """
        scheme, netloc, path, params, query, fragment = urlparse(self._config.engine_url)

        for part in paths:
            path = PATH_SEP.join((path.rstrip(PATH_SEP), part.lstrip(PATH_SEP)))

        query = urlencode({**dict(parse_qsl(query)), **query_params})

        return urlunparse((scheme, netloc, path, params, query, fragment))

    # ########################################################################
    # Бизнес-логика
    # ########################################################################
    def process_external_tasks(self, *topics, max_tasks=10) -> Iterable:
        """Получить и обработать доступные внешние задачи.

        max_tasks - ограничение по одновременному количеству задач.
        """

        topics = topics or self.get_external_tasks_topics()
        tasks = self.fetch_and_lock_external_tasks(*topics, max_tasks=max_tasks)
        with ThreadPoolExecutor() as executor:
            yield from executor.map(
                self.process_external_task, tasks
            )

    def process_external_task(self, task: model.ExternalTask) -> Any:
        """Обработать внешнюю задачу.

        Делегируется обработчику, назначенному на топик задачи.
        """

        try:
            handler = self._topic_registry.get(task.topic)
            return handler(task, self)
        except Exception as error:  # pylint: disable=broad-except
            logger.exception('Ошибка обработки задачи %s: %s', task.id, str(error))

    # ########################################################################
    # Прокси-методы регистрации обработчиков
    # ########################################################################
    def register_topic_handler(
        self,
        topic: model.Topic,
        handler: TopicHandler
    ) -> None:
        """Зарегистрировать обработчик топика."""

        return self._topic_registry.register(topic, handler)

    def replace_topic_handler(
        self,
        topic: model.Topic,
        handler: TopicHandler
    ) -> None:
        """Заменить обработчик топика."""

        return self._topic_registry.replace(topic, handler)

    def available(self) -> bool:
        """Проверяет доступность движка."""

        try:
            return self._session.get(
                self._build_url('/engine')
            ).status_code == 200
        except requests.exceptions.RequestException:
            return False
